/*
 *  Program to process the Jargon File (v 2.9.11+)
 *  Gives fortunes, searches for keywords etc..
 *
 * 	Copyright (c) 1992-1999 Kai Risku <krisku@iki.fi>
 *
 *  You may copy and use this program freely as long as this
 *  notice is left intact. This program is provided "AS IS"
 *  without warranty of any kind. The copyright owner may not
 *  be held liable for any damages, direct or consequential, 
 *  which may result from the the use of this program.
 */

#include <sys/stat.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <time.h>
#include <stdlib.h>
#ifdef __MSDOS__
#include <io.h>
#else
#include <ctype.h>
#endif

#define VER "2.4"
#define JARG_LIST    1
#define JARG_DISPLAY 2
#define IDX_FILE	"/usr/local/lib/jarg.idx"

void make_index( char *indexfile, char *textfile );
void get_quote( char *indexfile, char *opt );
void search_word( char *indexfile, char *keyword );
void read_quote( long offset , int length , char *filename );
void disp_entries( char *indexfile, char *string, int mode );

#ifdef __MSDOS__
#  define _srand srand
#  define _rand rand
#else
char *_fullpath( char *dest, char *src, int maxlen );
char *strlwr( char *str );
#  define stricmp strcasecmp
#  define _srand srand48
#  define _rand lrand48
#endif


int main( int argc , char *argv[] )
{
    int i;
    char indxfile[80], str[80];
#ifdef __MSDOS__
    char *p;
#endif

    if (argc==1)
    {
        puts("Jargon tool ver "VER", "__DATE__" by Kai Risku\n\n"
             "Usage: jarg [option] [arguments]\n\n"
             "jarg -i jargonfile    - creates index for jargonfile (full pathname)\n"
             "jarg -f [maxlines]    - give a random jargon fortune\n"
             "jarg -l foo           - lists all keywords containing 'foo'\n"
             "jarg -d bar           - displays all entries whose keywords contain 'bar'\n"
             "jarg [-] baz          - displays entry for keyword 'baz'\n"
             );

        exit ( 0 );
    }

#ifdef __MSDOS__
    strcpy( indxfile, argv[0] );
    p = strchr ( indxfile, '.' );
    if ( p!=NULL )
        *p = '\0';
    strcat( indxfile, ".idx" );
#else
    strcpy( indxfile, IDX_FILE );
#endif

    if ( argv[1][0]=='-' )
    {
        switch ( tolower(argv[1][1]) )
	{
	  case 'i':   
	    make_index ( indxfile, argv[2] );
	    break;
	  case 'f':   
	    get_quote ( indxfile, argv[2] );
	    break;

	  case 'l':
	  case 'd':   
	    strcpy ( str, "" );
	    for ( i=2; i<argc; i++ )
	    {
		strcat ( str, argv[i] );
		strcat ( str, " " );
	    }
	    if ( str[strlen(str)-1]==' ' )
		str[strlen(str)-1] = '\0';
	    switch( tolower( argv[1][1] ) )
	    {
	      case 'l': 
		disp_entries( indxfile, str, JARG_LIST );
		break;
	      case 'd':
		disp_entries( indxfile, str, JARG_DISPLAY );
		break;
	    }
	    break;
	    
	  case 0:  
	    strcpy ( str, ":" );
	    for ( i=2; i<argc; i++ )
	    {
		strcat ( str, argv[i] );
		strcat ( str, " " );
	    }
	    str[strlen(str)-1] = ':';
	    search_word ( indxfile, str );
	    break;

	  case '-':
	    break;
	    
	  default:    
	    fprintf ( stderr, "Unknown option: %s\n", argv[1] );
	    break;
	}
    }
    else
    {
	strcpy ( str, ":" );
        for ( i=1; i<argc; i++ )
	{
            strcat ( str, argv[i] );
            strcat ( str, " " );
	}
        str[strlen(str)-1] = ':';
        search_word ( indxfile, str );
    }
    
    return ( 0 );
}


void make_index( char *indexfile, char *textfile )
{
    int maxqtes, lines, isentry;
    FILE *fp, *idx;
    char text[256], keyword[80], *p;
    long filepos, fpos2;

    idx = fopen ( indexfile, "w" );
    if ( idx==NULL )
    {
        fprintf( stderr , "Couldn't open file %s for writing.\n" , indexfile );
        return;
    }

    fp = fopen ( textfile, "rb" );
    if( fp==NULL )
    {
        fprintf( stderr , "Couldn't open file %s for reading.\n" , textfile );
        fclose( idx );
	unlink ( indexfile );
        return;
    }

    if ( _fullpath ( keyword, textfile, 79 )==NULL )
    {
	fprintf ( stderr, "Illegal pathname: %s\n", textfile );
	fclose ( idx );
	unlink ( indexfile );
	return;
    }
    maxqtes = lines = isentry = 0;
    fprintf ( idx, "%s,00000\n", keyword );
    printf( "\nCreating indexfile %s for file %s\n" , indexfile, keyword );

    fpos2 = filepos = 0;
    while ( fgets(text,sizeof(text),fp)!=NULL &&
	    strncmp(text,"The Jargon Lexicon",18) )
        /* skip all introductory texts */ ;

    while ( fgets(text,sizeof(text),fp)!=NULL /* &&
	    strncmp(text,":(Lexicon Entries End Here):",28) */ )
    {
        p = text;
        while( *p==' ' || *p=='\t' )
            p++;
        if ( (p[0]==':' && p[1]>=' ') ||
	     (p[0]=='=' && p[1]>=' ' && lines>1) )
        {
            if ( isentry )
            {
                fprintf ( idx, ":%s:%ld,%d\n", keyword, filepos, lines );
                maxqtes++;
                isentry = 0;
            }

            if ( p[0]==':' )
            {
                filepos = fpos2;
                lines = 1;
                isentry = 1;
                sscanf ( p, ":%[^:]", keyword );
            }
        }
	else if ( text[0]=='*' && lines==1 )
	{
	    isentry = 0;
	}
        else
        {
            if ( isentry )
                lines++;
        }
        fpos2 = ftell(fp);
    }
    if ( isentry )
    {
        fprintf ( idx, ":%s:%ld,%d\n", keyword, filepos, lines );
        maxqtes++;
    }

    fclose( idx );
    fclose( fp );
    printf( "%d keywords indexed.\n" , maxqtes );
    idx = fopen ( indexfile, "r+" );
    _fullpath ( keyword, textfile, 79 );
    fprintf ( idx, "%s=%05d\n", keyword, maxqtes );
    fclose ( idx );
}


void get_quote( char *indexfile, char *opt )
{
    int qtes, quoteno, lines, maxlines;
    long quotepos;
    char text[256], filename[80], *p;
    FILE *fp;

    maxlines = 32767;
    if ( opt!=NULL )
        maxlines = atoi ( opt ) + 1;
    if ( maxlines<2 )
    {
        fprintf ( stderr, "Maxlines must be a positive integer!\n" );
        return;
    }

    fp = fopen ( indexfile, "r" );
    if ( fp==NULL )
    {
        fprintf ( stderr, "Could not open indexfile %s\n", indexfile );
        return;
    }
    fgets ( text, sizeof(text), fp );
    sscanf ( text, "%[^=]=%d", filename, &qtes );
    if ( qtes<1 )
    {
        fprintf ( stderr, "Error in indexfile!\n" );
        fclose ( fp );
        return;
    }

    _srand( time(NULL) );

    do
    {
        rewind ( fp );
        quoteno = _rand() % qtes + 2;
        while ( quoteno>0 )
        {
            fgets ( text, sizeof(text), fp );
            quoteno--;
        }
        p = strrchr ( text, ':' );
        sscanf ( p, ":%ld,%d", &quotepos, &lines );
    } while ( lines>maxlines );

    fclose ( fp );

    read_quote( quotepos, lines, filename );
}


void read_quote( long offset, int length, char *filename )
{
    FILE *fp;
    char quote[256];
    int i;

    fp = fopen ( filename, "r" );
    if ( fp==NULL )
        return;

    fseek( fp, offset, 0 );

    for ( i=0; i<length; i++ )
    {
        fgets ( quote, sizeof(quote), fp );
        printf ( "%s", quote );
    }

    fclose ( fp );
}


void search_word ( char *indexfile, char *keyword )
{
    FILE *fp;
    int qtes, lines;
    long quotepos;
    char text[256], filename[80], *p;

    fp = fopen ( indexfile, "r" );
    if ( fp==NULL )
    {
        fprintf ( stderr, "Could not open indexfile %s\n", indexfile );
        return;
    }

    fgets ( text, sizeof(text), fp );
    sscanf ( text, "%[^=]=%d", filename, &qtes );
    if ( qtes<1 )
    {
        fprintf ( stderr, "Error in indexfile!\n" );
        fclose ( fp );
        return;
    }

    while ( fgets ( text, sizeof(text), fp )!=NULL )
    {
        p = strrchr ( text, ':' );
        sscanf ( p, ":%ld,%d", &quotepos, &lines );
        p[1] = '\0';

        if ( stricmp(text,keyword)==0 )
            read_quote( quotepos, lines, filename );
    }

    fclose ( fp );
}


void disp_entries ( char *indexfile, char *string, int mode )
{
    FILE *fp;
    int qtes, lines;
    long quotepos;
    char text[256], str[256], filename[256], *p;

    strlwr ( string );
    fp = fopen ( indexfile, "r" );
    if ( fp==NULL )
    {
        fprintf ( stderr, "Could not open indexfile %s\n", indexfile );
        return;
    }

    fgets ( text, sizeof(text), fp );
    if( sscanf ( text, "%[^=]=%d", filename, &qtes )!=2 )
    {
        fprintf ( stderr, "Error in indexfile!\n" );
        fclose ( fp );
        return;
    }

    while ( fgets ( text, sizeof(text), fp )!=NULL )
    {
        p = strrchr ( text, ':' );
        sscanf ( p, ":%ld,%d", &quotepos, &lines );
        *p = '\0';

	strcpy( str, text );
	strlwr ( str );
        if( strstr( str+1, string )!=NULL )
        {
	    switch( mode )
	    {
	      case JARG_LIST:
		printf ( "%s:\n", text );
		break;
	      case JARG_DISPLAY:
		read_quote( quotepos, lines, filename );
		break;
	    }
        }
    }

    fclose ( fp );
}

#ifndef __MSDOS__
char *_fullpath ( char *dest, char *src, int maxlen )
{
    if ( src[0]=='/' )
    {
	sprintf ( dest, "%-.*s", maxlen, src );
	return ( dest );
    }
    return ( NULL );
}

char *strlwr( char *str )
{
    int i;
    
    for( i=0; str[i]; i++ )
	str[i] = tolower( str[i] );

    return( str );
}

#endif /* !__MSDOS__ */
